'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

# From https://salsa.debian.org/nm-team/nm-templates
NM_TEMPLATES = {
        # https://salsa.debian.org/nm-team/nm-templates/-/blob/master/nm_assigned.txt?ref_type=heads 
        'assigned': '''To get started, please tell me about yourself, how you
            came to GNU/Linux and free software, and why you want to volunteer
            your time. Please describe the contributions you have made to
            Debian, your primary areas of interest within Debian, and any goals
            you wish to accomplish.''',

        # https://salsa.debian.org/nm-team/nm-templates/-/blob/master/nm_pp1.txt?ref_type=heads
        'pp1.PH0': '''First, please have a careful read of the Social Contract
            and the DFSG. What do you think are their main points?''',
        'pp1.PH1': '''What is Debian's approach to non-free software? Why?
            Is non-free part of the Debian System? Please also explain the
            difference between non-free and the other sections.''',
        'pp1.PH2': '''Suppose that Debian were offered a Debian-specific
            license to package a certain piece of software: would we put it in
            main?''',
        'pp1.PH3': '''Please explain the difference between free as in "free
            speech" and as in "free beer". Is Debian mainly about free as in
            "free speech" or free as in "free beer"?''',
        'pp1.PH4': '''What is your opinion about how the DFSG should be applied
            to files that are not software, such as documentation, firmware,
            images, etc? Specifically, DFSG section 2 "Source Code".''',
        'pp1.PH5': '''How do you check if a license is DFSG-compatible? Who
            has the final say over what can be included in Debian?''',
        'pp1.PH6': '''There are a few "tests" for this purpose, based on (not
            really) common situations. Explain them to me and point out which
            common problems can be discovered by them.''',
        'pp1.PH7': '''At https://people.debian.org/~joerg/bad.licenses.tar.bz2
            you can find a tarball of bad licenses. Please compare the graphviz
            and three other (your choice) licenses with the first nine points
            of the DFSG and show a few examples of where they do not comply
            with the DFSG. There's no need to compare word for word (which
            would be impossible for some licenses anyway), but you should spot
            the biggest mistakes.
            Note: the graphviz license is bad for the brain: don't take too
            much of it.
            Also note: graphviz and qmail have now changed licenses in favour
            of free ones.''',
        'pp1.PHa': '''Are there any sections of the DFSG or Social Contract
            that you might like to see changed? If so, which ones, and why?''',
        # https://salsa.debian.org/nm-team/nm-templates/-/blob/master/nm_pp1_extras.txt?ref_type=heads
        'pp1e.PH0': '''First, please explain the key points of the Social
            Contract and the DFSG _in your own words_. Also, describe what you
            think about these documents. Please don't just rephrase each point
            of the Social Contract and the DFSG, but try to summarize it as
            a whole.''',
        'pp1e.PH8': '''Donald Knuth, author of TeX, insists that no-one has
            the right to modify the source code of TeX, and that any changes
            must be made using "change files" (a sort of patch file). Is this
            allowed for a program for the main component of Debian?''',
        'pp1e.PH9': '''The GNU Free Documentation License (GFDL) has been
            heavily discussed on debian-legal in the past, leading to a GR [6].
            Please summarize the problems Debian has with the GFDL, and what
            you think about the GR outcome.
            [6] https://www.debian.org/vote/2006/vote_001''',
        'pp1e.PHb': '''An upstream released the first version of his program
            under GPL-2+ license. A year later he decided to release the version
            2.0 of the same program under BSD-3-Clause. Is there any licensing
            violation here? Can this software be packaged in Debian?''',
        # https://salsa.debian.org/nm-team/nm-templates/-/blob/master/nm_pp2.txt?ref_type=heads
        'pp2.BT2': '''You have a package (current packaged version is 1.3-1),
            with a set of open bugs:

              #123: eats lots of memory (fixed in 1.2)
              #345: please package version 1.4
              #567: typo in package description
              #789: segfaults every Sunday (fixed in 1.4)
              #901: please move config file to /usr/local/etc

            Please write a changelog entry for the upload of the new version
            1.4 of the package.  What do you do with the bugs not mentioned
            in the changelog?''',
        'pp2.BT6': '''What do you do if you want to reach the submitter of a
            bug and keep a copy of the mail in the BTS?''',
        'pp2.BT8': '''Please explain what a usertag is and why they're useful.
            How do you tag a bug with a usertag?''',
}



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import List, Union, Callable, Any
import time
import sys
import argparse
import numpy as np
import functools as ft
from . import defaults

console = defaults.console


def retry_ratelimit(func: Callable,
                    exception: Exception,
                    retry_interval: int = 15) -> Callable:
    '''
    A decorator to retry the function call when exception occurs.

    OpenAI API doc provides some other methods to retry:
    https://platform.openai.com/docs/guides/rate-limits/error-mitigation

    Args:
        func (Callable): The function to be retried.
        exception (Exception): The exception to catch and retry upon.
        retry_interval (int): The interval in seconds to wait before retrying.

    Returns:
        Callable: A wrapped function with retry logic.
    '''

    @ft.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        while True:
            try:
                result = func(*args, **kwargs)
                break
            except exception:  # pragma: no cover
                console.log(
                    f'Rate limit reached. Will retry after {retry_interval} seconds.'
                )
                time.sleep(retry_interval)
        return result

    return wrapper


class AbstractEmbeddingModel(object):
    '''
    Abstract class for embedding models.
    '''

    # the model name
    model: str = 'none'

    # the embedding dimension (after reduction)
    dim: int = 0

    def __init__(self) -> None:  # pragma: no cover
        pass

    def embed(self, text: str) -> np.ndarray:  # pragma: no cover
        '''
        Embed a single text string.

        Args:
            text (str): The text to embed.

        Returns:
            np.ndarray: The embedding vector.
        '''
        raise NotImplementedError('This is an abstract method.')

    def batch_embed(self, texts: List[str]) -> np.ndarray:  # pragma: no cover
        '''
        Embed a batch of text strings.

        Args:
            texts (List[str]): List of texts to embed.

        Returns:
            np.ndarray: A matrix of embedding vectors (normalized).
        '''
        raise NotImplementedError('This is an abstract method.')

    def __call__(self, text: Union[str, List[str]]) -> np.ndarray:
        '''
        Call method to embed text or batch of texts.

        Args:
            text (Union[str, List[str]]): Text or list of texts to embed.

        Returns:
            np.ndarray: The embedding vector or matrix (normalized).
        '''
        if isinstance(text, str):
            return self.embed(text)
        elif isinstance(text, list):
            return self.batch_embed(text)
        else:  # pragma: no cover
            raise ValueError('Invalid input type.')


class RandomEmbedding(AbstractEmbeddingModel):
    '''
    Random embedding model for testing purposes.
    '''

    def __init__(self, args: object = None) -> None:
        self.model = 'random'
        self.dim = args.embedding_dim

    def embed(self, text: str) -> np.ndarray:
        '''
        Embed a single text string using random vectors.

        Args:
            text (str): The text to embed.

        Returns:
            np.ndarray: The embedding vector.
        '''
        vector = np.random.randn(self.dim)
        vector = vector / np.linalg.norm(vector)
        return vector

    def batch_embed(self, texts: List[str]) -> np.ndarray:
        '''
        Embed a batch of text strings using random vectors.

        Args:
            texts (List[str]): List of texts to embed.

        Returns:
            np.ndarray: A matrix of embedding vectors.
        '''
        matrix = np.random.randn(len(texts), self.dim)
        matrix = matrix / np.linalg.norm(matrix, axis=1)[:, np.newaxis]
        return matrix


class OpenAIEmbedding(AbstractEmbeddingModel):
    '''
    OpenAI embedding model implementation.
    '''

    def __init__(self, args: object = None) -> None:
        from openai import OpenAI
        self.client = OpenAI(api_key=args.openai_api_key,
                             base_url=args.openai_base_url)
        self.model = args.openai_embedding_model
        self.dim = args.embedding_dim

    def embed(self, text: str) -> np.ndarray:
        '''
        Embed a single text string using OpenAI.

        Args:
            text (str): The text to embed.

        Returns:
            np.ndarray: The embedding vector.
        '''
        from openai import RateLimitError
        func = retry_ratelimit(self.client.embeddings.create, RateLimitError)
        response = func(input=text, model=self.model, dimensions=self.dim)
        vector = np.array(response.data[0].embedding)
        vector = vector / np.linalg.norm(vector)
        return vector

    def batch_embed(self, texts: List[str]) -> np.ndarray:
        '''
        Embed a batch of text strings using OpenAI.

        Args:
            texts (List[str]): List of texts to embed.

        Returns:
            np.ndarray: A matrix of embedding vectors.
        '''
        from openai import RateLimitError
        func = retry_ratelimit(self.client.embeddings.create, RateLimitError)
        response = func(input=texts, model=self.model)
        matrix = np.stack([x.embedding for x in response.data])[:, :self.dim]
        matrix = matrix / np.linalg.norm(matrix, axis=1)[:, np.newaxis]
        return matrix


class GoogleEmbedding(AbstractEmbeddingModel):
    '''
    Google embedding model implementation.

    Example model: "models/text-embedding-004"
    This model has a maximum dimension of 768.
    Example dimension: 256

    Reference:
    https://github.com/google-gemini/cookbook/blob/main/quickstarts/Embeddings.ipynb
    '''

    def __init__(self, args: object = None) -> None:
        import google.generativeai as genai
        genai.configure(api_key=args.google_api_key)
        self.client = genai
        self.model = args.google_embedding_model
        self.dim = args.embedding_dim

    def embed(self, text: str) -> np.ndarray:
        '''
        Embed a single text string using Gemini.

        Args:
            text (str): The text to embed.

        Returns:
            np.ndarray: The embedding vector.
        '''
        from google.api_core.exceptions import ResourceExhausted
        func = retry_ratelimit(self.client.embed_content, ResourceExhausted)
        response = func(model=self.model,
                        content=text,
                        output_dimensionality=self.dim)
        vector = np.array(response['embedding'])
        vector = vector / np.linalg.norm(vector)
        return vector

    def batch_embed(self, texts: List[str]) -> np.ndarray:
        '''
        Embed a batch of text strings using Gemini.

        Args:
            texts (List[str]): List of texts to embed.

        Returns:
            np.ndarray: A matrix of embedding vectors.
        '''
        from google.api_core.exceptions import ResourceExhausted
        func = retry_ratelimit(self.client.embed_content, ResourceExhausted)
        response = func(model=self.model,
                        content=texts,
                        output_dimensionality=self.dim)
        matrix = np.stack(response['embedding'])[:, :self.dim]
        matrix = matrix / np.linalg.norm(matrix, axis=1)[:, np.newaxis]
        return matrix


def get_embedding_model(
        args: object) -> AbstractEmbeddingModel:  # pragma: no cover
    '''
    Get the embedding model based on the provided arguments.

    Args:
        args (object): The arguments containing model configuration.

    Returns:
        AbstractEmbeddingModel: The instantiated embedding model.
    '''
    if args.embedding_frontend == 'openai':
        return OpenAIEmbedding(args)
    elif args.embedding_frontend == 'google':
        return GoogleEmbedding(args)
    elif args.embedding_frontend == 'random':
        return RandomEmbedding(args)
    else:  # pragma: no cover
        raise ValueError('Invalid embedding frontend.')


def main(argv: List[str]) -> None:
    '''
    Main function to parse arguments and perform embedding.

    Args:
        argv (List[str]): Command-line arguments.
    '''
    conf = defaults.Config()
    parser = argparse.ArgumentParser()
    parser.add_argument('text',
                        type=str,
                        nargs='?',
                        default='Your text string goes here',
                        help='Text to embed')
    args = parser.parse_args(argv)

    model = get_embedding_model(conf)
    vector = model.embed(args.text)
    print('vector.shape:', vector.shape)
    print('vector.min:', vector.min())
    print('vector.max:', vector.max())
    print('vector.mean:', vector.mean())
    print('vector.std:', vector.std())
    print('vector[:10]:', vector[:10])
    print('vector[-10:]:', vector[-10:])


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:])



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
__version__ = '0.7.6'
__copyright__ = '2024-2025, Mo Zhou <lumin@debian.org>'
__license__ = 'LGPL-3.0+'
# do not load all components.
# some components like llm, and backend, requires much more dependencies


def version() -> None:
    print(f'DebGPT {__version__}')
    print(f'Copyright {__copyright__}')
    print(f'Released under {__license__} license.')



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import Optional
import textwrap
import sys
from typing import Union, List, Tuple
import sqlite3
import argparse
import numpy as np
import lz4.frame
from .defaults import console


class VectorDB:
    '''
    A class to manage a database of vectors using SQLite.

    Attributes:
        __dtype (np.dtype): The default data type for vectors.
        connection (sqlite3.Connection): The SQLite connection object.
        cursor (sqlite3.Cursor): The SQLite cursor object.
        dim (int): The dimension of the vectors.
    '''

    __dtype = np.float32

    def __init__(self, db_name: str = 'VectorDB.sqlite', dimension: int = 256):
        '''
        Initialize a VectorDB object. It is suggested to use a file name that
        contains both the embedding model and the embedding size to avoid
        errors.

        Args:
            db_name (str): The name of the database file. Defaults to 'VectorDB.sqlite'.
            dimension (int): The dimension of the vectors. Defaults to 256.
        '''
        console.log('Connecting to database:', db_name)
        self.connection: sqlite3.Connection = sqlite3.connect(db_name)
        self.cursor: sqlite3.Cursor = self.connection.cursor()
        self._create_table()
        self.dim: int = dimension

    def _create_table(self) -> None:
        '''
        Create the vectors table in the database if it doesn't exist.
        '''
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS vectors (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                source TEXT NOT NULL,
                text BLOB NOT NULL,
                vector BLOB NOT NULL
            )
        ''')
        self.connection.commit()

    def add(self, source: str, text: str, vector: Union[list,
                                                        np.ndarray]) -> None:
        '''
        Add a vector to the database. The vector is normalized before storage.

        Args:
            source (str): The source of the vector, e.g., file path, URL.
            text (str): The original text content corresponding to the vector.
            vector (Union[list, np.ndarray]): The vector to store.
        '''
        assert len(vector) >= self.dim
        vector_np: np.ndarray = np.array(vector, dtype=self.__dtype)
        vector_np_reduction: np.ndarray = vector_np[:self.dim]
        vector_np_reduction = vector_np_reduction / np.linalg.norm(
            vector_np_reduction)
        vector_bytes: bytes = vector_np_reduction.tobytes()
        text_compressed: bytes = lz4.frame.compress(text.encode())
        self.cursor.execute(
            'INSERT INTO vectors (source, text, vector) VALUES (?, ?, ?)', (
                source,
                text_compressed,
                vector_bytes,
            ))
        self.connection.commit()

    def _decode_row(self, row: List) -> List[Union[int, str, np.ndarray]]:
        '''
        Decode a row from the database into its original components.

        Args:
            row (List): The row from the database.

        Returns:
            List[Union[int, str, np.ndarray]]: The decoded components.
        '''
        idx, source, text_compressed, vector_bytes = row
        vector_np: np.ndarray = np.frombuffer(vector_bytes, dtype=self.__dtype)
        text_uncompressed: str = lz4.frame.decompress(text_compressed).decode()
        return [idx, source, text_uncompressed, vector_np]

    def get_byid(self, vector_id: int) -> List[Union[int, str, np.ndarray]]:
        '''
        Retrieve a row from the database by its ID.

        Args:
            vector_id (int): The ID of the vector to retrieve.

        Returns:
            List[Union[int, str, np.ndarray]]: The retrieved vector and its metadata.

        Raises:
            ValueError: If the vector with the specified ID is not found.
        '''
        self.cursor.execute('SELECT * FROM vectors WHERE id = ?',
                            (vector_id, ))
        result: Tuple = self.cursor.fetchone()
        if result:
            return self._decode_row(result)
        raise ValueError(f'Vector with id={vector_id} not found')

    def __getitem__(self, vector_id: int) -> List[Union[int, str, np.ndarray]]:
        '''
        Retrieve a row from the database by its ID using the index operator.

        Args:
            vector_id (int): The ID of the vector to retrieve.

        Returns:
            List[Union[int, str, np.ndarray]]: The retrieved vector and its metadata.

        Raises:
            ValueError: If the vector with the specified ID is not found.
        '''
        return self.get_byid(vector_id)

    def get_all(self) -> List[List[Union[int, str, np.ndarray]]]:
        '''
        Retrieve all rows from the vectors table.

        Returns:
            List[List[Union[int, str, np.ndarray]]]: All rows from the table.
        '''
        self.cursor.execute('SELECT * FROM vectors')
        results: List[Tuple] = self.cursor.fetchall()
        return [self._decode_row(row) for row in results]

    def as_array(self) -> Tuple[np.ndarray, np.ndarray]:
        '''
        Retrieve all IDs and vectors from the database as numpy arrays.

        Returns:
            Tuple[np.ndarray, np.ndarray]: Arrays of vector IDs and vectors.
        '''
        self.cursor.execute('SELECT id, vector FROM vectors')
        results: List[Tuple[int, bytes]] = self.cursor.fetchall()
        results = [(idx, np.frombuffer(vector, dtype=self.__dtype))
                   for idx, vector in results]
        # repack them into numpy arrays
        idxs, vectors = list(zip(*results))
        idxs_array: np.ndarray = np.array(idxs)
        matrix: np.ndarray = np.stack(vectors)
        return idxs_array, matrix

    def delete_byid(self, vector_id: int) -> None:
        '''
        Delete a vector from the database by its ID.

        Args:
            vector_id (int): The ID of the vector to delete.
        '''
        self.cursor.execute('DELETE FROM vectors WHERE id = ?', (vector_id, ))
        self.connection.commit()

    def close(self) -> None:
        '''
        Close the database connection.
        '''
        self.connection.close()

    def retrieve(self,
                 vector: np.ndarray,
                 topk: int = 3) -> List[List[Union[float, str]]]:
        '''
        Retrieve the nearest vectors from the database based on cosine similarity.

        Args:
            vector (np.ndarray): The vector to compare against.
            topk (int): The number of nearest vectors to retrieve. Defaults to 3.

        Returns:
            List[List[Union[float, str]]]: A list of the nearest vectors and their metadata.
        '''
        idxs, matrix = self.as_array()
        assert matrix.ndim == 2
        assert vector.ndim == 1
        vector = vector / np.linalg.norm(vector)
        cosine: np.ndarray = (matrix @ vector.reshape(-1, 1)).flatten()
        argsort: np.ndarray = np.argsort(cosine)[::-1][:topk]
        documents: List[List[Union[float, str]]] = []
        for idx, sim in zip(idxs[argsort], cosine[argsort]):
            _, source, text, _, = self.get_byid(int(idx))
            doc: List[Union[float, str]] = [sim, source, text]
            documents.append(doc)
        return documents

    def ls(self,
           id: Optional[int] = None
           ) -> List[List[Union[int, str, np.ndarray]]]:
        '''
        List all vectors in the database.

        Returns:
            List[List[Union[int, str, np.ndarray]]]: All vectors and their metadata.
        '''
        if id is not None:
            vectors: List[List[Union[int, str,
                                     np.ndarray]]] = [self.get_byid(id)]
        else:
            vectors: List[List[Union[int, str, np.ndarray]]] = self.get_all()
        for v in vectors:
            idx, source, text, vector = v
            console.print(
                f'id[{idx:5d}]',
                f'len(vector)={len(vector)},',
                f'len(text)={len(text):5d}',
                f'source={repr(source)},',
                f'text={textwrap.shorten(text, 32)}',
            )
        return vectors

    def show(self, idx: int) -> None:
        '''
        Show the vector with the given index.

        Args:
            idx (int): The index of the vector to show.
        '''
        vector: List[Union[int, str, np.ndarray]] = self.get_byid(idx)
        idx, source, text, vector = vector
        print(
            f'[{idx:4d}]',
            f'source={repr(source)},',
            f'text={repr(text)}',
            f'\nlen(vector)={len(vector)}',
            f'\nvector={vector}',
        )


def main(argv: List[str]) -> None:
    '''
    Main function to handle command-line interface for the VectorDB.

    Args:
        argv (List[str]): Command-line arguments.
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument('--db',
                        type=str,
                        default='VectorDB.sqlite',
                        help='Database file name')
    subparsers = parser.add_subparsers(dest='action')
    _ = subparsers.add_parser('demo')
    parser_ls = subparsers.add_parser('ls')
    parser_ls.add_argument('id',
                           type=int,
                           default=None,
                           nargs='?',
                           help='ID of the vector to list')
    parser_show = subparsers.add_parser('show')
    parser_show.add_argument('id', type=int, help='ID of the vector to show')
    parser_rm = subparsers.add_parser('rm')
    parser_rm.add_argument('id', type=int, help='ID of the vector to remove')
    args = parser.parse_args(argv)

    if args.action == 'ls':
        db = VectorDB(args.db)
        db.ls(args.id)
        db.close()
    elif args.action == 'show':
        db = VectorDB(args.db)
        db.show(args.id)
        db.close()
    elif args.action == 'rm':
        db = VectorDB(args.db)
        db.delete_byid(args.id)
        console.log(f'Deleted vector with id={args.id}')
        db.close()
    elif args.action == 'demo':
        db = VectorDB(args.db)
        for i in range(10):
            v: np.ndarray = np.random.rand(256)
            db.add(f'vector_{i}', str(v), v)
        db.add('ones', str(np.ones(256)), np.ones(256))
        db.close()
    else:
        parser.print_help()


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:])



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import List, Dict, Union, Optional
import argparse
import os
import json
import uuid
import sys
import time
import functools as ft
import shlex

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.styles import Style
from rich.console import Console, Group
from rich.live import Live
from rich.status import Status
from rich.markdown import Markdown
from rich.markup import escape
from rich.text import Text
from rich.padding import Padding
from rich.style import Style as richStyle

from . import defaults

console = defaults.console
console_stdout = Console()


def _check(messages: List[Dict]):
    '''
    communitation protocol.
    both huggingface transformers and openapi api use this
    '''
    assert isinstance(messages, list)
    assert all(isinstance(x, dict) for x in messages)
    assert all('role' in x.keys() for x in messages)
    assert all('content' in x.keys() for x in messages)
    assert all(isinstance(x['role'], str) for x in messages)
    assert all(isinstance(x['content'], str) for x in messages)
    assert all(x['role'] in ('system', 'user', 'assistant') for x in messages)


def retry_ratelimit(func: callable,
                    exception: Exception,
                    retry_interval: int = 15):
    '''
    a decorator to retry the function call when exception occurs.

    OpenAI API doc provides some other methods to retry:
    https://platform.openai.com/docs/guides/rate-limits/error-mitigation
    '''

    @ft.wraps(func)
    def wrapper(*args, **kwargs):
        while True:
            try:
                result = func(*args, **kwargs)
                break
            except exception:
                console.log(
                    f'Rate limit reached. Will retry after {retry_interval} seconds.'
                )
                time.sleep(15)
        return result

    return wrapper


class AbstractFrontend():
    '''
    The frontend instance holds the whole chat session. The context is the whole
    session for the next LLM query. Historical chats is also a part of the
    context for following up questions. You may feel LLMs smart when they
    get information from the historical chat in the same session.
    '''

    NAME = 'AbstractFrontend'

    def __init__(self, args):
        self.uuid = uuid.uuid4()
        self.session = []
        self.debgpt_home = args.debgpt_home
        self.monochrome = args.monochrome
        self.multiline = args.multiline
        self.render_markdown = args.render_markdown
        console.log(f'{self.NAME}> Starting conversation {self.uuid}')

    def reset(self):
        '''
        clear the context. No need to change UUID I think.
        '''
        self.session = []

    def oneshot(self, message: str) -> str:
        '''
        Generate response text from the given question, without history.
        And do not print anything. Just return the response text silently.

        Args:
            message: a string, the question.
        Returns:
            a string, the response text.
        '''
        raise NotImplementedError('please override AbstractFrontend.oneshot()')

    def query(self, messages: List[Dict]) -> str:
        '''
        Generate response text from the given chat history. This function
        will also handle printing and rendering.

        Args:
            messages: a list of dict, each dict contains a message.
        Returns:
            a string, the response text.
        the messages format can be found in _check(...) function above.
        '''
        raise NotImplementedError('please override AbstractFrontend.query()')

    def update_session(self, messages: Union[List, Dict, str]) -> None:
        if isinstance(messages, list):
            # reset the chat with provided message list
            self.session = messages
        elif isinstance(messages, dict):
            # just append a new dict
            self.session.append(messages)
        elif isinstance(messages, str):
            # just append a new user dict
            self.session.append({'role': 'user', 'content': messages})
        else:
            raise TypeError(type(messages))
        _check(self.session)

    def __call__(self, *args, **kwargs):
        try:
            res = self.query(*args, **kwargs)
            return res
        except Exception as e:
            # this will only appear in dumped session files
            self.update_session({'role': 'system', 'content': str(e)})
            raise e

    def dump(self):
        fpath = os.path.join(self.debgpt_home, str(self.uuid) + '.json')
        with open(fpath, 'wt') as f:
            json.dump(self.session, f, indent=2)
        console.log(f'{self.NAME}> Conversation saved at {fpath}')

    def __len__(self):
        '''
        Calculate the number of messages from user and assistant in the session,
        excluding system message.
        '''
        return len([x for x in self.session if x['role'] != 'system'])


class EchoFrontend(AbstractFrontend):
    '''
    A frontend that echoes the input text. Don't worry, this is just for
    running unit tests.
    '''
    NAME = 'EchoFrontend'
    lossy_mode: bool = False
    lossy_rate: int = 2

    def __init__(self, args: Optional[object] = None):
        # do not call super().__init__(args) here.
        self.session = []
        self.stream = False
        self.monochrome = False
        self.multiline = False
        self.render_markdown = False

    def oneshot(self, message: str) -> str:
        if self.lossy_mode:
            return message[::self.lossy_rate]
        else:
            return message

    def query(self, messages: Union[List, Dict, str]) -> list:
        self.update_session(messages)
        new_input = self.session[-1]['content']
        if self.lossy_mode:
            response = new_input[::self.lossy_rate]
        else:
            response = new_input
        console_stdout.print(response)
        new_message = {'role': 'assistant', 'content': response}
        self.update_session(new_message)
        return self.session[-1]['content']

    def dump(self):
        pass


class OpenAIFrontend(AbstractFrontend):
    '''
    https://platform.openai.com/docs/quickstart?context=python
    '''
    NAME: str = 'OpenAIFrontend'
    debug: bool = False
    stream: bool = True
    system_message: str = defaults.OPENAI_SYSTEM_MESSAGE

    def __init__(self, args):
        super().__init__(args)
        try:
            from openai import OpenAI
        except ImportError:
            console.log('please install OpenAI package: "pip install openai"')
            exit(1)
        self.client = OpenAI(api_key=args.openai_api_key,
                             base_url=args.openai_base_url)
        self.model = args.openai_model
        # XXX: some models do not support system messages yet. nor temperature.
        if self.model not in ('o1-mini', 'o1-preview'):
            self.session.append({"role": "system", "content": self.system_message})
            self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        else:
            self.kwargs = {}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')

    def oneshot(self, message: str) -> str:

        def _func() -> str:
            _callable = self.client.chat.completions.create
            completions = _callable(model=self.model,
                                    messages=[{
                                        "role": "user",
                                        "content": message
                                    }],
                                    **self.kwargs)
            return completions.choices[0].message.content

        from openai import RateLimitError
        return retry_ratelimit(_func, RateLimitError)()

    def query(self, messages: Union[List, Dict, str]) -> list:
        # add the message into the session
        self.update_session(messages)
        if self.debug:
            console.log('send:', self.session[-1])
        completion = self.client.chat.completions.create(model=self.model,
                                                         messages=self.session,
                                                         stream=self.stream,
                                                         **self.kwargs)
        if self.stream:
            think, chunks = [], []
            cursor = chunks
            if self.render_markdown:
                with Live(Markdown('')) as live:
                    for chunk in completion:
                        if chunk.choices[0].delta.content is None:
                            continue
                        piece = chunk.choices[0].delta.content
                        if piece == '</think>' and think is not None:
                            cursor = chunks
                        elif piece == '<think>':
                            cursor = think
                        else:
                            cursor.append(piece)
                        # join chunks
                        buffer_think = ''.join(think)
                        part1 = Text(buffer_think)
                        part1 = Padding(part1, (0, 2),
                                        style=richStyle(dim=True, italic=True))
                        buffer_chunk = ''.join(chunks)
                        part2 = Markdown(buffer_chunk)
                        group = Group(part1, part2)
                        live.update(group, refresh=True)
            else:
                for chunk in completion:
                    if chunk.choices[0].delta.content is None:
                        continue
                    piece = chunk.choices[0].delta.content
                    chunks.append(piece)
                    print(piece, end="", flush=True)
            generated_text = ''.join(chunks)
            if not generated_text.endswith('\n'):
                print()
                sys.stdout.flush()
        else:
            generated_text = completion.choices[0].message.content
            if self.render_markdown:
                console_stdout.print(Markdown(generated_text))
            else:
                console_stdout.print(escape(generated_text))
        new_message = {'role': 'assistant', 'content': generated_text}
        self.update_session(new_message)
        if self.debug:
            console.log('recv:', self.session[-1])
        return self.session[-1]['content']


class AnthropicFrontend(AbstractFrontend):
    '''
    https://docs.anthropic.com/en/api/getting-started
    But we are currently using OpenAI API.

    The max_token limit for each model can be found here:
    https://docs.anthropic.com/en/docs/about-claude/models
    '''
    NAME = 'AnthropicFrontend'
    debug: bool = False
    stream: bool = True
    max_tokens: int = 4096

    def __init__(self, args):
        super().__init__(args)
        try:
            from anthropic import Anthropic
        except ImportError:
            console.log(
                'please install Anthropic package: "pip install anthropic"')
            exit(1)
        self.client = Anthropic(api_key=args.anthropic_api_key,
                                base_url=args.anthropic_base_url)
        self.model = args.anthropic_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')

    def oneshot(self, message: str) -> str:

        def _func():
            _callable = self.client.messages.create
            completion = _callable(model=self.model,
                                   messages=[{
                                       "role": "user",
                                       "content": message
                                   }],
                                   max_tokens=self.max_tokens,
                                   **self.kwargs)
            return completion.content[0].text

        from anthropic import RateLimitError
        return retry_ratelimit(_func, RateLimitError)()

    def query(self, messages: Union[List, Dict, str]) -> list:
        # add the message into the session
        self.update_session(messages)
        if self.debug:
            console.log('send:', self.session[-1])
        if self.stream:
            chunks = []
            with self.client.messages.stream(model=self.model,
                                             messages=self.session,
                                             max_tokens=self.max_tokens,
                                             **self.kwargs) as stream:
                if self.render_markdown:
                    with Live(Markdown('')) as live:
                        for chunk in stream.text_stream:
                            chunks.append(chunk)
                            live.update(Markdown(''.join(chunks)),
                                        refresh=True)
                else:
                    for chunk in stream.text_stream:
                        chunks.append(chunk)
                        print(chunk, end="", flush=True)
            generated_text = ''.join(chunks)
            if not generated_text.endswith('\n'):
                print()
                sys.stdout.flush()
        else:
            completion = self.client.messages.create(
                model=self.model,
                messages=self.session,
                max_tokens=self.max_tokens,
                stream=self.stream,
                **self.kwargs)
            generated_text = completion.content[0].text
            if self.render_markdown:
                console_stdout.print(Markdown(generated_text))
            else:
                console_stdout.print(escape(generated_text))
        new_message = {'role': 'assistant', 'content': generated_text}
        self.update_session(new_message)
        if self.debug:
            console.log('recv:', self.session[-1])
        return self.session[-1]['content']


class GoogleFrontend(AbstractFrontend):
    '''
    https://ai.google.dev/gemini-api/docs
    '''
    NAME = 'GoogleFrontend'
    debug: bool = False
    stream: bool = True

    def __init__(self, args):
        super().__init__(args)
        try:
            import google.generativeai as genai
        except ImportError:
            console.log(
                'please install gemini package: "pip install google-generativeai"'
            )
            exit(1)
        genai.configure(api_key=args.google_api_key)
        self.client = genai.GenerativeModel(args.google_model)
        self.chat = self.client.start_chat()
        self.kwargs = genai.types.GenerationConfig(
            temperature=args.temperature, top_p=args.top_p)
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(args.google_model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')

    def oneshot(self, message: str, *, retry: bool = True) -> str:

        def _func():
            _callable = self.client.generate_content
            result = _callable(message, generation_config=self.kwargs)
            return result.text

        from google.api_core.exceptions import ResourceExhausted
        return retry_ratelimit(_func, ResourceExhausted)()

    def query(self, messages: Union[List, Dict, str]) -> list:
        # add the message into the session
        self.update_session(messages)
        if self.debug:
            console.log('send:', self.session[-1])
        if self.stream:
            chunks = []
            response = self.chat.send_message(self.session[-1]['content'],
                                              stream=True,
                                              generation_config=self.kwargs)
            if self.render_markdown:
                with Live(Markdown('')) as live:
                    for chunk in response:
                        chunks.append(chunk.text)
                        live.update(Markdown(''.join(chunks)), refresh=True)
            else:
                for chunk in response:
                    chunks.append(chunk.text)
                    print(chunk.text, end="", flush=True)
            generated_text = ''.join(chunks)
        else:
            response = self.chat.send_message(self.session[-1]['content'],
                                              generation_config=self.kwargs)
            generated_text = response.text
            if self.render_markdown:
                console_stdout.print(Markdown(generated_text))
            else:
                console_stdout.print(escape(generated_text))
        new_message = {'role': 'assistant', 'content': generated_text}
        self.update_session(new_message)
        if self.debug:
            console.log('recv:', self.session[-1])
        return self.session[-1]['content']


class XAIFrontend(OpenAIFrontend):
    '''
    https://console.x.ai/
    '''
    NAME = 'xAIFrontend'

    def __init__(self, args):
        super().__init__(args)
        from openai import OpenAI
        self.client = OpenAI(api_key=args.xai_api_key,
                             base_url='https://api.x.ai/v1/')
        self.session.append({"role": "system", "content": self.system_message})
        self.model = args.xai_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class NvidiaFrontend(OpenAIFrontend):
    '''
    This is a frontend for Nvidia's NIM/NeMo service.
    https://build.nvidia.com/
    '''
    NAME = 'Nvidia-Frontend'

    def __init__(self, args):
        super().__init__(args)
        from openai import OpenAI
        self.client = OpenAI(api_key=args.nvidia_api_key,
                             base_url=args.nvidia_base_url)
        self.session.append({"role": "system", "content": self.system_message})
        self.model = args.nvidia_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class LlamafileFrontend(OpenAIFrontend):
    '''
    https://github.com/Mozilla-Ocho/llamafile
    '''
    NAME = 'LlamafileFrontend'

    def __init__(self, args):
        AbstractFrontend.__init__(self, args)
        from openai import OpenAI
        self.client = OpenAI(api_key='no-key-required',
                             base_url=args.llamafile_base_url)
        self.session.append({"role": "system", "content": self.system_message})
        self.model = 'llamafile from https://github.com/Mozilla-Ocho/llamafile'
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class OllamaFrontend(OpenAIFrontend):
    '''
    https://github.com/ollama/ollama
    '''
    NAME = 'OllamaFrontend'

    def __init__(self, args):
        AbstractFrontend.__init__(self, args)
        from openai import OpenAI
        self.client = OpenAI(api_key='no-key-required',
                             base_url=args.ollama_base_url)
        self.session.append({"role": "system", "content": self.system_message})
        self.model = args.ollama_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class LlamacppFrontend(OpenAIFrontend):
    '''
    https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md
    '''
    NAME = 'LlamacppFrontend'

    def __init__(self, args):
        AbstractFrontend.__init__(self, args)
        from openai import OpenAI
        self.client = OpenAI(api_key='no-key-required',
                             base_url=args.llamacpp_base_url)
        self.session.append({"role": "system", "content": self.system_message})
        self.model = 'model-is-specified-at-the-llama-server-arguments'
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class DeepSeekFrontend(OpenAIFrontend):
    '''
    https://api-docs.deepseek.com/
    '''
    NAME = 'DeepSeekFrontend'

    def __init__(self, args):
        AbstractFrontend.__init__(self, args)
        from openai import OpenAI
        self.client = OpenAI(api_key=args.deepseek_api_key,
                             base_url=args.deepseek_base_url)
        if args.deepseek_model not in ('deepseek-reasoner'):
            # see the usage recommendations at
            # https://huggingface.co/deepseek-ai/DeepSeek-R1
            self.session.append({"role": "system", "content": self.system_message})
        self.model = args.deepseek_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class vLLMFrontend(OpenAIFrontend):
    '''
    https://docs.vllm.ai/en/stable/serving/openai_compatible_server.html
    '''
    NAME = 'vLLMFrontend'

    def __init__(self, args):
        AbstractFrontend.__init__(self, args)
        from openai import OpenAI
        self.client = OpenAI(api_key='your-vllm-api-key',
                             base_url=args.vllm_base_url)
        self.session.append({"role": "system", "content": self.system_message})
        self.model = args.vllm_model
        self.kwargs = {'temperature': args.temperature, 'top_p': args.top_p}
        if args.verbose:
            console.log(f'{self.NAME}> model={repr(self.model)}, ' +
                        f'temperature={args.temperature}, top_p={args.top_p}.')


class ZMQFrontend(AbstractFrontend):
    '''
    ZMQ frontend communicates with a self-hosted ZMQ backend.
    '''
    NAME = 'ZMQFrontend'
    debug: bool = False
    stream: bool = False

    def __init__(self, args):
        import zmq
        super().__init__(args)
        self.zmq_backend = args.zmq_backend
        self.socket = zmq.Context().socket(zmq.REQ)
        self.socket.connect(self.zmq_backend)
        console.log(
            f'{self.NAME}> Connected to ZMQ backend {self.zmq_backend}.')
        #
        if hasattr(args, 'temperature'):
            console.log(
                'warning! --temperature not yet supported for this frontend')
        if hasattr(args, 'top_p'):
            console.log('warning! --top_p not yet supported for this frontend')

    def query(self, content: Union[List, Dict, str]) -> list:
        if isinstance(content, list):
            self.session = content
        elif isinstance(content, dict):
            self.session.append(content)
        elif isinstance(content, str):
            self.session.append({'role': 'user', 'content': content})
        _check(self.session)
        msg_json = json.dumps(self.session)
        if self.debug:
            console.log('send:', msg_json)
        self.socket.send_string(msg_json)
        msg = self.socket.recv()
        self.session = json.loads(msg)
        _check(self.session)
        if self.debug:
            console.log('recv:', self.session[-1])
        return self.session[-1]['content']


def get_username():
    try:
        import getpass
        return getpass.getuser()
    except:
        pass
    try:
        import pwd
        return pwd.getpwuid(os.getuid())[0]
    except:
        pass
    try:
        return os.getlogin()
    except:
        pass
    # common shell env
    for env in ('USER', 'USERNAME', 'LOGNAME'):
        if u := os.environ.get(env):
            return u
    # final fallback
    return 'user'


def create_frontend(args):
    if args.frontend == 'zmq':
        frontend = ZMQFrontend(args)
    elif args.frontend == 'openai':
        frontend = OpenAIFrontend(args)
    elif args.frontend == 'anthropic':
        frontend = AnthropicFrontend(args)
    elif args.frontend == 'google':
        frontend = GoogleFrontend(args)
    elif args.frontend == 'xai':
        frontend = XAIFrontend(args)
    elif args.frontend == 'nvidia':
        frontend = NvidiaFrontend(args)
    elif args.frontend == 'llamafile':
        frontend = LlamafileFrontend(args)
    elif args.frontend == 'ollama':
        frontend = OllamaFrontend(args)
    elif args.frontend == 'llamacpp':
        frontend = LlamacppFrontend(args)
    elif args.frontend == 'deepseek':
        frontend = DeepSeekFrontend(args)
    elif args.frontend == 'vllm':
        frontend = vLLMFrontend(args)
    elif args.frontend == 'dryrun':
        frontend = None
    elif args.frontend == 'echo':
        frontend = EchoFrontend(args)
    else:
        raise NotImplementedError
    return frontend


def interact_once(f: AbstractFrontend, text: str) -> None:
    '''
    we have prepared text -- let frontend send it to LLM, and this function
    will print the LLM reply.

    f: any frontend instance from the current source file.
    text: the text to be sent to LLM.
    '''
    if f.stream:
        end = '' if not f.render_markdown else '\n'
        if f.monochrome:
            lprompt = escape(f'LLM[{2+len(f)}]> ')
            console.print(lprompt, end=end, highlight=False, markup=False)
        else:
            lprompt = f'[bold green]LLM[{2+len(f)}]>[/bold green] '
            console.print(lprompt, end=end)
        _ = f(text)
    else:
        with Status('LLM', spinner='line'):
            _ = f(text)


def interact_with(f: AbstractFrontend) -> None:
    # create prompt_toolkit style
    if f.monochrome:
        prompt_style = Style([('prompt', 'bold')])
    else:
        prompt_style = Style([('prompt', 'bold fg:ansibrightcyan'),
                              ('', 'bold ansiwhite')])

    # Completer with several keywords keywords to be completed
    class CustomCompleter(Completer):

        def get_completions(self, document, complete_event):
            # Get the current text before the cursor
            text_before_cursor = document.text_before_cursor

            # Check if the text starts with '/'
            if text_before_cursor.startswith('/'):
                # Define the available keywords
                keywords = ['/quit', '/save', '/reset']

                # Generate completions for each keyword
                for keyword in keywords:
                    if keyword.startswith(text_before_cursor):
                        yield Completion(keyword, -len(text_before_cursor))

    # start prompt session
    prompt_session = PromptSession(style=prompt_style,
                                   multiline=f.multiline,
                                   completer=CustomCompleter())

    # if multiline is enabled, print additional help message
    if f.multiline:
        console.print(
            'In multiline mode, please press [Meta+Enter], or [Esc] followed by [Enter] to send the message.'
        )

    # loop
    user = get_username()
    try:
        while text := prompt_session.prompt(
                f'{user}[{1+len(f)}]> '):
            # parse escaped interaction commands
            if text.startswith('/'):
                cmd = shlex.split(text)
                if cmd[0] == '/save':
                    # save the last LLM reply to a file
                    if len(cmd) != 2:
                        console.print('syntax error: /save <path>')
                        continue
                    path = cmd[-1]
                    with open(path, 'wt') as fp:
                        fp.write(f.session[-1]['content'])
                    console.log(f'The last LLM response is saved at {path}')
                elif cmd[0] == '/reset':
                    if len(cmd) != 1:
                        console.print('syntax error: /reset')
                        continue
                    f.reset()
                elif cmd[0] == '/quit':
                    if len(cmd) != 1:
                        console.print('syntax error: /quit')
                        continue
                    break
                else:
                    console.print(f'unknown command: {cmd[0]}')
            else:
                interact_once(f, text)
    except EOFError:
        pass
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    ag = argparse.ArgumentParser()
    ag.add_argument('--zmq_backend', '-B', default='tcp://localhost:11177')
    ag.add_argument('--frontend',
                    '-F',
                    default='zmq',
                    choices=('dryrun', 'zmq', 'openai', 'anthropic', 'google',
                             'llamafile', 'ollama', 'vllm'))
    ag.add_argument('--debgpt_home', default=os.path.expanduser('~/.debgpt'))
    ag = ag.parse_args()
    console.print(ag)

    frontend = create_frontend(ag)
    f = frontend
    import IPython
    IPython.embed(colors='neutral')



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import os
import pytest
from debgpt.policy import DebianPolicy
from debgpt.policy import DebianDevref


def test_policy(tmpdir: str) -> None:
    """
    Test the DebianPolicy class by checking specific sections.

    Args:
        tmpdir (str): The temporary directory to use for testing.
    """
    policy = DebianPolicy()
    # specific section of the policy, indexed by section number (as string)
    for section in ('1', '4.6', '4.9.1'):
        text = policy[section]
    # Convert the entire policy to a string
    whole = str(policy)
    assert len(whole) > 1000

    # test __iter__ and __next__
    for i, section in enumerate(policy):
        assert isinstance(section, str)
        assert len(section) > 0

    # test __len__
    assert len(policy) > 0

    # test __getitem__ with int
    for i in range(len(policy)):
        text = policy[i]
        assert len(text) > 0


@pytest.mark.parametrize('section', ('2', '2.1', '3.1.1'))
def test_devref(tmpdir: str, section: str) -> None:
    """
    Test the DebianDevref class by checking specific sections.

    Args:
        tmpdir (str): The temporary directory to use for testing.
        section (str): The section of the Debian Developer's Reference to test.
    """
    devref = DebianDevref()
    # Print the specific section of the developer's reference
    print(devref[section])
    # Convert the entire developer's reference to a string
    whole = str(devref)
    # Assert that the entire developer's reference string is longer than 1000 characters
    assert len(whole) > 1000



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import List, Union, Dict, Tuple
import pytest
from debgpt import mapreduce
from debgpt import frontend
from debgpt import reader
import os
import time
import numpy as np
import sys
import io
import itertools as it


@pytest.fixture
def frtnd():
    f = frontend.EchoFrontend()
    f.lossy_mode = True
    return f


@pytest.fixture
def chunk():
    return reader.Entry(
        path='path',
        content='content',
        wrapfun=lambda x: '{spec}: {content}'.format(spec='spec', content=x),
        wrapfun_chunk=lambda x: '{spec} (lines {start}-{end}): {content}'.
        format(spec='spec', content=x, start=0, end=-1),
    )


def test_shorten():
    string = 'a b c d e f g h i j k l m n o p q r s t u v w x y z' * 1000
    assert len(mapreduce.shorten(string)) <= 100


def test_pad_chunk_before_map(chunk):
    question = 'test question'
    filled_template = mapreduce.pad_chunk_before_map(chunk, question)
    assert isinstance(filled_template, str)
    assert filled_template
    assert len(filled_template) > len(chunk)


def test_map_chunk(chunk, frtnd):
    question = 'test question'
    result = mapreduce.map_chunk(chunk, question, frtnd, verbose=True)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


def test_map_serial(chunk, frtnd):
    chunks = [chunk] * 10
    question = 'test question'
    results = mapreduce.map_serial(chunks, question, frtnd)
    assert isinstance(results, list)
    assert len(results) == 10
    for result in results:
        assert isinstance(result, str)
        assert result
        assert len(result) > 0


def test_map_parallel(chunk, frtnd):
    chunks = [chunk] * 10
    question = 'test question'
    results = mapreduce.map_parallel(chunks, question, frtnd)
    assert isinstance(results, list)
    assert len(results) == 10
    for result in results:
        assert isinstance(result, str)
        assert result
        assert len(result) > 0


def test_pad_two_results_for_reduce():
    results = ['a', 'b']
    question = 'test question'
    filled_template = mapreduce.pad_two_results_for_reduce(*results, question)
    assert isinstance(filled_template, str)
    assert filled_template
    assert len(filled_template) > 0


def test_reduce_two_chunks(frtnd):
    question = 'test question'
    result = mapreduce.reduce_two_chunks('a',
                                         'b',
                                         question,
                                         frtnd,
                                         verbose=True)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


@pytest.mark.parametrize('repeat', range(1, 10))
def test_pad_many_results_for_reduce(repeat, frtnd):
    results = ['a'] * repeat
    question = 'test question'
    filled_template = mapreduce.pad_many_results_for_reduce(results, question)
    assert isinstance(filled_template, str)
    assert filled_template
    assert len(filled_template) > 0


@pytest.mark.parametrize('repeat', range(1, 10))
def test_reduce_many_chunks(repeat, frtnd):
    chunks = ['a'] * repeat
    question = 'test question'
    result = mapreduce.reduce_many_chunks(chunks,
                                          question,
                                          frtnd,
                                          verbose=True)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


def test_group_strings_by_length():
    strings = 'a b c d e f g h i j k l m n o p q r s t u v w x y z'.split()
    groups = mapreduce.group_strings_by_length(strings, 2)
    assert len(groups) == 13
    for group in groups:
        assert len(group) == 2


def test_reduce_serial(frtnd):
    chunks = ['abc'] * 100
    question = 'test question'
    result = mapreduce.reduce_serial(chunks, question, frtnd, verbose=True)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


def test_reduce_serial_compact(frtnd):
    chunks = ['abc'] * 100
    question = 'test question'
    result = mapreduce.reduce_serial_compact(chunks,
                                             question,
                                             frtnd,
                                             verbose=True,
                                             max_chunk_size=20)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


def test_reduce_parallel(frtnd):
    chunks = ['abc'] * 100
    question = 'test question'
    result = mapreduce.reduce_parallel(chunks, question, frtnd, verbose=True)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


def test_reduce_parallel_compact(frtnd):
    chunks = ['abc'] * 100
    question = 'test question'
    result = mapreduce.reduce_parallel_compact(chunks,
                                               question,
                                               frtnd,
                                               verbose=True,
                                               max_chunk_size=20)
    assert isinstance(result, str)
    assert result
    assert len(result) > 0


@pytest.mark.parametrize(
    'parallel,compact_map,compact_reduce,repeat,max_chunk_size',
    it.product([1, 2, 4], [True, False], [True, False], [1, 100], [20, 100]))
def test_mapreduce_super_long_context(tmpdir, frtnd, parallel, compact_map,
                                      compact_reduce, repeat, max_chunk_size):
    text = ['a b c d e f g h i j k l m n o p q r s t u v w x y z'] * repeat
    text = '\n'.join(text)
    with open(tmpdir / 'test.txt', 'wt') as f:
        f.write(text)
    spec = tmpdir.join('test.txt').strpath

    # do mapreduce
    aggregated = mapreduce.mapreduce_super_long_context(
        spec=spec,
        max_chunk_size=max_chunk_size,
        frtnd=frtnd,
        user_question='test question',
        verbose=True,
        compact_map_mode=compact_map,
        compact_reduce_mode=compact_reduce,
        parallelism=parallel,
    )
    assert isinstance(aggregated, str)
    assert aggregated
    assert len(aggregated) > 0



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import Iterable, Optional, Dict
import urwid
import os
import re
import sys
from . import defaults
from . import arguments

default = defaults.Config()
console = defaults.console

_TITLE = 'DebGPT Configurator'


def _abort_on_None(value: Optional[str]) -> None:
    if value is None:
        print('Aborted.')
        exit(1)


class ListBoxVimKeys(urwid.ListBox):

    def keypress(self, size, key) -> None:
        key_map = {
            'k': 'up',
            'j': 'down',
        }
        if key in ('esc', 'q', 'Q'):
            sys.exit(1)
        super().keypress(size, key_map.get(key, key))


class SingleChoice(object):

    # we store the user choice here
    _choice: Optional[str] = None

    @staticmethod
    def exit_on_esc(key: str) -> None:
        if key in ('esc', ):
            sys.exit(1)

    def item_chosen(self, button: urwid.Button, choice: str) -> None:
        SingleChoice._choice = choice
        raise urwid.ExitMainLoop(choice)

    def __init__(self,
                 title: str,
                 question: str,
                 choices: Iterable[str],
                 helpmsg: str,
                 statusmsg: str,
                 *,
                 focus: int = 0):
        # header
        header = urwid.AttrMap(urwid.Text(title, align='center'), 'reversed')
        footer = urwid.Text(statusmsg)
        # build the question between header and menu
        body = [
            urwid.Divider(),
            urwid.Text(question),
            urwid.Divider(),
        ]
        # build the menu
        buttons = []
        for c in choices:
            button = urwid.Button(c)
            urwid.connect_signal(button, "click", self.item_chosen, c)
            buttons.append(urwid.AttrMap(button, None, focus_map="reversed"))
        pile = urwid.Pile(buttons)
        pile.set_focus(focus)
        body.append(urwid.LineBox(pile))
        # build the help message between menu and footer
        body.extend([
            urwid.Divider(),
            urwid.Text(helpmsg),
        ])
        # assemble the widgets
        body = ListBoxVimKeys(urwid.SimpleFocusListWalker(body))
        frame = urwid.Frame(header=header, body=body, footer=footer)
        frame = urwid.Padding(frame, left=1, right=1)
        loop = urwid.MainLoop(frame,
                              palette=[("reversed", "standout", "")],
                              unhandled_input=self.exit_on_esc)
        self.loop = loop

    def run(self):
        try:
            self.loop.run()
        except urwid.ExitMainLoop:
            return None
        return self._choice


class SingleEdit(object):

    # we store the user choice here
    _choice: Optional[str] = None

    @staticmethod
    def exit_on_esc(key: str) -> None:
        if key in ('esc', ):
            sys.exit(1)
        elif key in ('enter', ):
            raise urwid.ExitMainLoop()

    def edit_update(self, edit: urwid.Edit, new_edit_text: str) -> None:
        self._choice = new_edit_text

    def __init__(self, title: str, question: str, default: str, helpmsg: str,
                 statusmsg: str):
        # process arguments
        self._choice = default
        # header and footer
        header = urwid.AttrMap(urwid.Text(title, align='center'), 'reversed')
        footer = urwid.Text(statusmsg)
        # build the question between header and menu
        body = [
            urwid.Divider(),
            urwid.Padding(urwid.Text(question), left=1, right=1),
            urwid.Divider(),
        ]
        # build the edit widget
        edit = urwid.Edit("", default)
        urwid.connect_signal(edit, 'change', self.edit_update)
        edit = urwid.LineBox(edit)
        body.append(edit)
        body.extend([
            urwid.Divider(),
            urwid.Padding(urwid.Text(helpmsg), left=1, right=1),
        ])
        # assemble the widgets
        body = urwid.ListBox(urwid.SimpleFocusListWalker(body))
        frame = urwid.Frame(header=header, body=body, footer=footer)
        frame = urwid.Padding(frame, left=1, right=1)
        loop = urwid.MainLoop(frame,
                              palette=[("reversed", "standout", "")],
                              unhandled_input=self.exit_on_esc)
        self.loop = loop

    def run(self):
        self.loop.run()
        return self._choice


def _request_frontend_specific_config(frontend: str,
                                      current_config: Dict = dict(),
                                      is_embedding: bool = False) -> dict:
    '''
    ask the user to provide the frontend-specific configuration
    '''
    conf = dict()

    # openai part
    if frontend == 'openai' and 'openai_base_url' not in current_config:
        value = SingleEdit(
            _TITLE, "Enter the OpenAI base url:", default['openai_base_url'],
            "Keep the default as is, if you do not intend to use this API on a different compatible service.",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['openai_base_url'] = value
    if frontend == 'openai' and 'openai_api_key' not in current_config:
        value = SingleEdit(
            _TITLE, "Enter the OpenAI API key:", default['openai_api_key'],
            "Typically your key can be found here: https://platform.openai.com/settings/organization/api-keys",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['openai_api_key'] = value
    if frontend == 'openai' and not is_embedding and 'openai_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the OpenAI API model name:",
            default['openai_model'],
            'If not sure, just keep the default. Available options: https://platform.openai.com/docs/models',
            'Press Enter to confirm. Press Esc to abort.').run()
        _abort_on_None(value)
        conf['openai_model'] = value
    if frontend == 'openai' and is_embedding and 'openai_embedding_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator",
            "Enter the OpenAI API embedding model name:",
            default['openai_embedding_model'],
            'If not sure, just keep the default.',
            'Press Enter to confirm. Press Esc to abort.').run()
        _abort_on_None(value)
        conf['openai_embedding_model'] = value

    # anthropic part
    if frontend == 'anthropic' and 'anthropic_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Anthropic API key:",
            default['anthropic_api_key'],
            "Typicall your key can be found here: https://console.anthropic.com/settings/keys",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['anthropic_api_key'] = value
    if frontend == 'anthropic' and 'anthropic_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Anthropic model name:",
            default['anthropic_model'],
            "If not sure, just keep the default. Available options: https://docs.anthropic.com/en/docs/about-claude/models",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['anthropic_model'] = value

    # google part
    if frontend == 'google' and 'google_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Google Gemini API key:",
            default['google_api_key'],
            "Typically found here: https://aistudio.google.com/app/apikey",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['google_api_key'] = value
    if frontend == 'google' and not is_embedding and 'google_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Google model name:",
            default['google_model'],
            "If not sure, just keep the default. Available options: https://ai.google.dev/gemini-api/docs/models/gemini",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['google_model'] = value
    if frontend == 'google' and is_embedding and 'google_embedding_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Google embedding model name:",
            default['google_embedding_model'],
            "If not sure, just keep the default.",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['google_embedding_model'] = value

    # xai part
    if frontend == 'xai' and 'xai_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the xAI API key:",
            default['xai_api_key'],
            "Typically found here: https://console.x.ai/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['xai_api_key'] = value
    if frontend == 'xai' and not is_embedding and 'xai_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the xAI model name:",
            default['xai_model'],
            "If not sure, just keep the default. Available options: https://console.x.ai/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['xai_model'] = value

    # nvidia part
    if frontend == 'nvidia' and 'nvidia_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the NVIDIA NIM/NeMo service url:",
            default['nvidia_base_url'],
            "Default is https://integrate.api.nvidia.com/v1",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['nvidia_base_url'] = value
    if frontend == 'nvidia' and 'nvidia_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the NVIDIA NIM/NeMo API key:",
            default['nvidia_api_key'],
            "Typically found here: https://build.nvidia.com/explore/discover",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['nvidia_api_key'] = value
    if frontend == 'nvidia' and not is_embedding and 'nvidia_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the NVIDIA NIM/NeMo model name:",
            default['nvidia_model'],
            "If not sure, just keep the default. Available options: https://build.nvidia.com/explore/discover",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['nvidia_model'] = value

    # ollama part
    if frontend == 'ollama' and 'ollama_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Ollama service url:",
            default['ollama_base_url'],
            "Reference: https://github.com/ollama/ollama/blob/main/README.md",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['ollama_base_url'] = value
    if frontend == 'ollama' and 'ollama_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the Ollama model name:",
            default['ollama_model'],
            "Reference: https://github.com/ollama/ollama/blob/main/README.md",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['ollama_model'] = value

    # llamacpp part
    if frontend == 'llamacpp' and 'llamacpp_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the llama-server (llama.cpp) service url:",
            default['llamacpp_base_url'],
            "Reference: https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['llamacpp_base_url'] = value

    # deepseek part
    if frontend == 'deepseek' and 'deepseek_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the DeepSeek service url:",
            default['deepseek_base_url'],
            "Just keep the default if you intend to use DeepSeek API service. Reference: https://api-docs.deepseek.com/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['deepseek_base_url'] = value

    if frontend == 'deepseek' and 'deepseek_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the DeepSeek API key:",
            default['deepseek_api_key'],
            "Typically found here: https://platform.deepseek.com/api_keys",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['deepseek_api_key'] = value

    if frontend == 'deepseek' and 'deepseek_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the DeepSeek model name:",
            default['deepseek_model'],
            "If not sure, just keep the default. As of Jan 24 2025, the available models are: deepseek-chat, deepseek-reasoner. The list of latest models can be found here: https://api-docs.deepseek.com/quick_start/pricing",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['deepseek_model'] = value

    # llamafile part
    if frontend == 'llamafile' and 'llamafile_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the LlamaFile service url:",
            default['llamafile_base_url'],
            "Reference: https://github.com/Mozilla-Ocho/llamafile",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['llamafile_base_url'] = value

    # vllm part
    if frontend == 'vllm' and 'vllm_base_url' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the vLLM service url:",
            default['vllm_base_url'],
            "Reference: https://docs.vllm.ai/en/stable/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['vllm_base_url'] = value
    if frontend == 'vllm' and 'vllm_api_key' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the vLLM API key:",
            default['vllm_api_key'],
            "Reference: https://docs.vllm.ai/en/stable/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['vllm_api_key'] = value
    if frontend == 'vllm' and 'vllm_model' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the vLLM model name:",
            default['vllm_model'],
            "Reference: https://docs.vllm.ai/en/stable/",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['vllm_model'] = value

    # zmq part
    if frontend == 'zmq' and 'zmq_backend' not in current_config:
        value = SingleEdit(
            "DebGPT Configurator", "Enter the DebGPT ZMQ Backend URL:",
            default['zmq_backend'],
            "The service endpoint where you launched debgpt backend.",
            "Press Enter to confirm. Press Esc to abort.").run()
        _abort_on_None(value)
        conf['zmq_backend'] = value

    # dryrun part
    if frontend == 'dryrun':
        pass

    return conf


def _request_common_cli_behavior_config() -> dict:
    '''
    ask the user to provide the common CLI behavior configuration
    '''
    conf = dict()
    # 1. whether to render LLM response markdown
    focus = {True: 0, False: 1}[default['render_markdown']]
    value = SingleChoice("DebGPT Configurator",
                         "Render LLM response (Markdown) in terminal?",
                         ['yes', 'no'],
                         "Default is 'yes' (recommended). This option \
produces fancy terminal printing with markdown stream.",
                         "Press Enter to confirm. Press Esc to abort.",
                         focus=focus).run()
    _abort_on_None(value)
    conf['render_markdown'] = value == 'yes'
    return conf


def _request_overwrite_config(dest: str) -> bool:
    '''
    ask the user whether to overwrite the existing configuration file
    '''
    value = SingleChoice(
        "DebGPT Configurator",
        f"Configuration file {repr(dest)} already exists. \
Overwrite?", ['no', 'yes'], "The existing options will be inherited. We will \
simply refresh the configuration file with the updates from this wizard. Press Esc to abort.",
        "Press Enter to confirm. Press Esc to abort.").run()
    _abort_on_None(value)
    return value == 'yes'


def _edit_config_template(config_template: str, key: str, value: str) -> str:
    '''
    Edit the configuration template (toml file). Updating the value of
    the specified key.

    Args:
        config_template: the configuration template as a string
        key: the key to be updated
        value: the value to be updated
    Returns:
        the updated configuration template
    '''
    lines = config_template.split('\n')
    newlines = []
    for line in lines:
        if re.match(r'^\s*{}\s*='.format(key), line):
            line = '{} = {}'.format(key, value)
        newlines.append(line)
    return '\n'.join(newlines)


def fresh_install_guide(dest: Optional[str] = None) -> dict:
    '''
    This function is a configuration guide for fresh installation of DebGPT.
    '''
    conf = dict()
    config_template: str = arguments.parse_args([]).config_template

    if dest and os.path.exists(dest):
        overwrite = _request_overwrite_config(dest)
        _abort_on_None(overwrite)
        if overwrite == False:
            print('Aborted.')
            exit(0)

    # step 1: select a frontend
    frontends = [
        'OpenAI    (GPT)    | commercial,  OpenAI-API',
        'Anthropic (Claude) | commercial,  Anthropic-API',
        'Google    (Gemini) | commercial,  Google-API',
        'xAI       (Grok)   | commercial,  OpenAI-API',
        'Nvidia    (*)      | commercial,  OpenAI-API', 
        'DeepSeek           | commercial,  OpenAI-API, MIT-Licensed Model',
        'Ollama    (*)      | self-hosted, OpenAI-API',
        'llama.cpp (*)      | self-hosted, OpenAI-API',
        'LlamaFile (*)      | self-hosted, OpenAI-API',
        'vLLM      (*)      | self-hosted, OpenAI-API',
        'ZMQ       (*)      | self-hosted, DebGPT built-in',
        'Dryrun    (N/A)    | debug,       DebGPT built-in',
    ]
    frontends_focus = {
        'openai': 0,
        'anthropic': 1,
        'google': 2,
        'xai': 3,
        'nvidia': 4,
        'deepseek': 5,
        'ollama': 6,
        'llamacpp': 7,
        'llamafile': 8,
        'vllm': 9,
        'zmq': 10,
        'dryrun': 11,
    }[default['frontend']]

    frontend = SingleChoice("DebGPT Configurator",
                            "Select a frontend that DebGPT will use:",
                            frontends,
                            "A frontend is a client that communicates with \
its corresponding backend that serves large language model (LLM). \
To use a commercial \
service, you may need to sign up and pay for an API key. Besides, \
if you have a spare GPU or a powerful CPU, you can take a look at the \
self-hosted LLM services. A web search can direct you to the details.\n\n\
This configurator will generate a minimal configuration file for you to \
make DebGPT work with the selected frontend.\n\n\
For advanced usages and more options, you may generate a configuration \
template with the following command for manual editing:\n\n\
  $ debgpt genconfig > ~/.debgpt/config.yaml\n\n\
This could be useful if you wish to switch among multiple frontends \
using the `--frontend|-F` argument.",
                            "Press Enter to confirm. Press Esc to abort.",
                            focus=frontends_focus).run()
    _abort_on_None(frontend)
    frontend = frontend.split(' ')[0].lower().replace('.', '')
    conf['frontend'] = frontend

    # step 2: ask for the frontend-specific configuration
    extra = _request_frontend_specific_config(frontend)
    conf.update(extra)

    # step 3: ask for the embedding frontend
    embedding_frontends = [
        'OpenAI    | commercial,  OpenAI-API',
        'Google    | commercial,  Google-API',
        'Random    | debug,       DebGPT built-in',
    ]
    embedding_frontends_focus = {
        'openai': 0,
        'google': 1,
        'random': 2,
    }[default['embedding_frontend']]
    embedding_frontend = SingleChoice(
        "DebGPT Configurator",
        "Select an embedding frontend that DebGPT will use:",
        embedding_frontends,
        "An embedding model turns text into vector embeddings, \
unlocking use cases like search. Choose a frontend that will compute the \
embedding vectors.\n\n\
The embedding frontend can be different from the frontend.\n\n\
If you are not going to use the embedding-realted feature, such as vectordb,\
retrieval, retrieval-augmented-generation (RAG), etc., you can select 'Random'.",
        "Press Enter to confirm. Press Esc to abort.",
        focus=embedding_frontends_focus).run()
    _abort_on_None(embedding_frontend)
    embedding_frontend = embedding_frontend.split(' ')[0].lower()
    conf['embedding_frontend'] = embedding_frontend

    # step 4: ask for the embedding frontend-specific configuration
    newconf = _request_frontend_specific_config(embedding_frontend,
                                                conf,
                                                is_embedding=True)
    conf.update(newconf)

    # step 3: ask for the common CLI behavior configuration
    extra = _request_common_cli_behavior_config()
    conf.update(extra)

    # edit the configuration template
    for k, v in conf.items():
        if isinstance(v, bool):
            v = 'true' if v else 'false'
            config_template = _edit_config_template(config_template, k, v)
        elif isinstance(v, str):
            config_template = _edit_config_template(config_template, k,
                                                    repr(v))
        else:
            raise ValueError('Unexpected type:', type(v))
    config_template += '\n'  # new line at the end

    # final: write configuration to specified destination
    if dest:
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        with open(dest, 'wt') as f:
            f.write(config_template)
        console.log('Config written to:', dest)
        console.print('[white on violet]>_< Enjoy DebGPT!')
    else:
        # verbose print
        console.print('Configuration (config.toml):')
        print('```')
        print(config_template)
        print('```')

    return conf


if __name__ == '__main__':
    miniconfig = fresh_install_guide()



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import pytest
from debgpt.cli import main

#@pytest.mark.parametrize('cmd', (
#    '-F dryrun',
#    '-F dryrun -Hf Makefile',
#    '-F dryrun -Qf Makefile',
#    '-F dryrun --tldr man',
#    '-F dryrun --man man',
#    '-F drurun --version',
#    '-F dryrun --cmd "man man"',
#))
#def test_cli_system_exit(cmd: str):
#    with pytest.raises(SystemExit):
#        main(cmd.split())



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from types import SimpleNamespace
import sys
import os
import numpy as np
import pytest
from debgpt import defaults
from debgpt import retrieval
from debgpt import reader


def test_vectorretriever_add(tmpdir):
    conf = defaults.Config()
    conf.db = os.path.join(tmpdir, 'test.db')
    embedding_frontend = conf.embedding_frontend
    api_key = conf[f'{embedding_frontend}_api_key']
    if api_key.startswith('your-') and api_key.endswith('-key'):
        pytest.skip(f'API Key for {embedding_frontend} not configured')
    retriever = retrieval.VectorRetriever(conf)
    # add some documents
    for i in range(2):
        retriever.add(f'temp{i}', f'fruit{i}')


def test_vectorretriever_retrieve_onfly(tmpdir):
    conf = defaults.Config()
    conf.db = os.path.join(tmpdir, 'test.db')
    embedding_frontend = conf.embedding_frontend
    api_key = conf[f'{embedding_frontend}_api_key']
    if api_key.startswith('your-') and api_key.endswith('-key'):
        pytest.skip(f'API Key for {embedding_frontend} not configured')
    retriever = retrieval.VectorRetriever(conf)
    # on-the-fly retrieval
    query = 'fruit'
    documents = ['fruit', 'sky', 'orange', 'dog', 'cat', 'apple', 'banana']
    results = retriever.retrieve_onfly(query, documents, topk=3)
    assert len(results) == 3
    for i, result in enumerate(results):
        score, source, text = result
        assert text in documents
        assert score >= 0.0 - 1e-5
        assert score <= 1.0 + 1e-5
        assert source is not None
        if i == 0:
            assert text == 'fruit'
            assert np.isclose(score, 1.0)
    print(results)


def test_vectorretriever_retrieve_from_db(tmpdir):
    conf = defaults.Config()
    conf.db = os.path.join(tmpdir, 'test.db')
    embedding_frontend = conf.embedding_frontend
    api_key = conf[f'{embedding_frontend}_api_key']
    if api_key.startswith('your-') and api_key.endswith('-key'):
        pytest.skip(f'API Key for {embedding_frontend} not configured')
    retriever = retrieval.VectorRetriever(conf)
    # insert some documents
    vectors = retriever.batch_add(
        ['temp'] * 7,
        ['fruit', 'sky', 'orange', 'dog', 'cat', 'apple', 'banana'])
    assert len(vectors) == 7
    # retrieve from db
    query = 'fruit'
    results = retriever.retrieve_from_db(query, topk=3)
    assert len(results) == 3
    for i, result in enumerate(results):
        score, source, text = result
        assert text in ['fruit', 'orange', 'apple', 'banana']
        assert score >= 0.0 - 1e-5
        assert score <= 1.0 + 1e-5
        assert source is not None
        if i == 0:
            assert text == 'fruit'
            assert np.isclose(score, 1.0)
    print(results)


def test_retrieval_main(tmpdir):
    common_args = ['--db', os.path.join(tmpdir, 'test.db'), '-E', 'random']
    retrieval.main([*common_args, 'add', 'x'])
    retrieval.main([*common_args, 'add', 'y'])
    retrieval.main([*common_args, 'add', 'z'])
    retrieval.main([*common_args, 'ret', 'w'])
    retrieval.main([*common_args, 'retrieve', 'w'])



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from types import SimpleNamespace
import sys
import os
import numpy as np
import pytest
from debgpt import defaults
from debgpt import frontend


@pytest.fixture
def conf() -> object:
    return defaults.Config()


def test_echo_frontend_oneshot(conf):
    f = frontend.EchoFrontend()
    assert f.oneshot('hello world') == 'hello world'


def test_echo_frontend_call(conf):
    f = frontend.EchoFrontend()
    assert f('hello world') == 'hello world'
    q = {'role': 'user', 'content': 'hello world'}
    assert f(q) == 'hello world'
    assert f([q]) == 'hello world'


def test_echo_frontend_query(conf):
    f = frontend.EchoFrontend()
    assert len(f.session) == 0
    assert f.query('hello world') == 'hello world'
    assert len(f.session) == 2
    assert f.query('hello world') == 'hello world'
    assert len(f.session) == 4
    assert f.query('hello world') == 'hello world'
    assert len(f.session) == 6



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from rich.status import Status
from typing import List, Dict
import argparse
import zmq
import torch as th
from . import llm
from .defaults import console


class AbstractBackend:

    def __init__(self, args):
        self.llm = llm.create_llm(args)

    def listen(self, args):
        raise NotImplementedError

    def server(self):
        raise NotImplementedError


def stat_messages(messages: List[Dict], llm):
    context_size = llm.tok.apply_chat_template(messages,
                                               tokenize=True,
                                               return_tensors='pt').size(1)
    ret = f'num_msgs={len(messages)}, ctx_size={context_size}; '
    ret += f'latest={messages[-1]}'
    return ret


class ZMQBackend(AbstractBackend):

    def __init__(self, args):
        super().__init__(args)
        self.socket = zmq.Context().socket(zmq.REP)
        binduri = args.host + ':' + str(args.port)
        self.socket.bind(binduri)
        console.log(f'ZMQBackend> bind URI {binduri}. Ready to serve.')

    def listen(self):
        while True:
            msg = self.socket.recv_json()
            yield msg

    def server(self):
        for query in self.listen():
            console.log(
                f'ZMQBackend> recv query: {stat_messages(query, self.llm)}',
                markup=False)
            with Status('LLM Calculating ...', spinner='line'):
                reply = self.llm(query)
            console.log(
                f'ZMQBackend> send reply: {stat_messages(reply, self.llm)}',
                markup=False)
            msg_json = zmq.utils.jsonapi.dumps(reply)
            self.socket.send(msg_json)


def create_backend(args):
    if args.backend_impl == 'zmq':
        backend = ZMQBackend(args)
    else:
        raise NotImplementedError(args.backend_impl)
    return backend


if __name__ == '__main__':
    ag = argparse.ArgumentParser()
    ag.add_argument('--port',
                    '-p',
                    type=int,
                    default=11177,
                    help='"11177" looks like "LLM"')
    ag.add_argument('--host', type=str, default='tcp://*')
    ag.add_argument('--backend_impl',
                    type=str,
                    default='zmq',
                    choices=('zmq', ))
    ag.add_argument('--max_new_tokens', type=int, default=512)
    ag.add_argument('--llm', type=str, default='Mistral7B')
    ag.add_argument('--device',
                    type=str,
                    default='cuda' if th.cuda.is_available() else 'cpu')
    ag.add_argument('--precision',
                    type=str,
                    default='fp16' if th.cuda.is_available() else '4bit')
    ag = ag.parse_args()
    console.log(ag)

    backend = create_backend(ag)
    try:
        backend.server()
    except KeyboardInterrupt:
        pass
    console.log('Server shut down.')



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import os
import pytest
import numpy as np
import tempfile
from debgpt.vectordb import VectorDB
from debgpt import vectordb


def _prepare_vdb(tmpdir: str, populate: bool = True) -> VectorDB:
    """
    Prepare a VectorDB instance with random vectors for testing.

    Args:
        tmpdir (str): Temporary directory path.
        populate (bool): Flag to indicate whether to populate the database with vectors.

    Returns:
        VectorDB: An instance of the VectorDB class.
    """
    # Create a temporary file for the database
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    vdb = VectorDB(os.path.join(tmpdir, temp_file.name))
    # Adding random vectors
    for i in range(10):
        v = np.random.rand(256)
        vdb.add(f'vector_{i}', str(v), v)
    # Add a constant vector for retrieval tests
    vdb.add(f'ones', str(np.ones(256)), np.ones(256))
    return vdb


def test_vectordb_init(tmpdir):
    """
    Test initialization of the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir, populate=False)
    assert vdb is not None
    vdb.close()


def test_vectordb_add(tmpdir):
    """
    Test adding vectors to the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    vdb.close()


def test_vectordb_get_byid(tmpdir):
    """
    Test retrieving a vector from the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    # Retrieve vector with index 11
    row = vdb.get_byid(11)
    idx, source, text, vector = row
    expected_vec = np.ones(256) / np.linalg.norm(np.ones(256))
    assert idx == 11
    assert source == 'ones'
    assert text == str(np.ones(256))
    assert np.allclose(vector, expected_vec)
    # Retrieve vector with index 11 using __index__ method
    row = vdb[11]
    idx, source, text, vector = row
    assert idx == 11
    assert source == 'ones'
    assert text == str(np.ones(256))
    assert np.allclose(vector, expected_vec)
    # Test retrieving a non-existent vector
    with pytest.raises(ValueError):
        vdb.get_byid(999)
    vdb.close()


def test_get_all(tmpdir):
    """
    Test retrieving all rows from the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    allrows = vdb.get_all()
    assert len(allrows) == 11
    # Validate each row
    for row in allrows:
        assert len(row) == 4
        assert isinstance(row[0], int)
        assert isinstance(row[1], str)
        assert isinstance(row[2], str)
        assert isinstance(row[3], np.ndarray)
    vdb.close()


def test_as_array(tmpdir):
    """
    Test retrieving all indices and vectors from the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    idx, matrix = vdb.as_array()
    assert len(idx) == 11
    assert isinstance(matrix, np.ndarray)
    assert matrix.shape == (11, 256)
    vdb.close()


def test_delete_byid(tmpdir):
    """
    Test deleting a vector from the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    allrows = vdb.get_all()
    assert len(allrows) == 11
    # Delete vector with index 1
    vdb.delete_byid(1)
    allrows = vdb.get_all()
    assert len(allrows) == 10
    vdb.close()


def test_retrieve(tmpdir):
    """
    Test retrieving similar vectors from the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    query_vector = np.ones(256) / np.linalg.norm(np.ones(256))
    documents = vdb.retrieve(query_vector, topk=3)
    assert len(documents) == 3
    for doc in documents:
        assert isinstance(doc, list)
        assert len(doc) == 3
        sim, source, text = doc
        assert isinstance(doc[0], float)
        assert isinstance(doc[1], str)
        assert isinstance(doc[2], str)
        if source == 'ones':
            assert np.isclose(sim, 1.0)
    vdb.close()


def test_vdb_ls(tmpdir):
    """
    Test listing vectors in the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    vdb.ls()
    vdb.close()


def test_vdb_show(tmpdir):
    """
    Test showing a specific vector in the VectorDB.
    """
    vdb = _prepare_vdb(tmpdir)
    vdb.show(1)
    vdb.close()


def test_main_demo(tmpdir):
    """
    Test running the demo command in the main function.
    """
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    path = os.path.join(tmpdir, temp_file.name)
    vectordb.main(['--db', path, 'demo'])


def test_main_ls(tmpdir):
    """
    Test running the list command in the main function.
    """
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    path = os.path.join(tmpdir, temp_file.name)
    vdb = VectorDB(os.path.join(tmpdir, temp_file.name))
    # Adding random vectors
    for i in range(10):
        v = np.random.rand(256)
        vdb.add(f'vector_{i}', str(v), v)
    vdb.close()
    vectordb.main(['--db', path, 'ls'])
    vectordb.main(['--db', path, 'ls', '1'])


def test_main_show(tmpdir):
    """
    Test running the show command in the main function.
    """
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    path = os.path.join(tmpdir, temp_file.name)
    vectordb.main(['--db', path, 'demo'])
    vectordb.main(['--db', path, 'show', '1'])
    # Test showing a non-existent vector
    with pytest.raises(ValueError):
        vectordb.main(['--db', path, 'show', '999'])


def test_main_rm(tmpdir):
    """
    Test running the remove command in the main function.
    """
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    path = os.path.join(tmpdir, temp_file.name)
    vectordb.main(['--db', path, 'demo'])
    vectordb.main(['--db', path, 'rm', '1'])


def test_main_help(tmpdir):
    """
    Test running the help command in the main function.
    """
    temp_file = tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False)
    path = os.path.join(tmpdir, temp_file.name)
    vectordb.main([])
    # Test the help command
    with pytest.raises(SystemExit):
        vectordb.main(['--help'])



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import List
import sys
from rich.markup import escape
from rich.panel import Panel
from rich.markdown import Markdown
import json
import argparse
from typing import Dict, Any
from .defaults import console


def process_entry(entry: Dict[str, Any], render: bool) -> None:
    """
    Processes a single chat entry and prints it to the console.

    Args:
        entry (Dict[str, Any]): A dictionary representing a chat entry with keys 'role' and 'content'.
        render (bool): A flag indicating whether to render assistant messages with rich Markdown.
    
    Raises:
        ValueError: If the entry role is unknown.
    """
    if entry['role'] == 'user':
        title = 'User Input'
        border_style = 'cyan'
        content = Panel(escape(entry['content']),
                        title=title,
                        border_style=border_style)
    elif entry['role'] == 'assistant':
        if render:
            content = Markdown(entry['content'])
        else:
            content = escape(entry['content'])
    elif entry['role'] == 'system':
        title = 'System Message'
        border_style = 'red'
        content = Panel(escape(entry['content']),
                        title=title,
                        border_style=border_style)
    else:
        raise ValueError(f'unknown role in {entry}')

    if render and entry['role'] != 'assistant':
        console.print(content)
    elif not render and entry['role'] == 'assistant':
        print(content)
    else:
        console.print(content)


def replay(path: str, render: bool = True) -> None:
    """
    Replays chat messages from a JSON file.

    Args:
        path (str): The file path to the JSON file containing chat messages.
        render (bool): A flag indicating whether to render assistant messages with rich Markdown.
    """
    with open(path) as f:
        J = json.load(f)

    for entry in J:
        process_entry(entry, render)


def main(argv: List[str] = sys.argv[1:]) -> None:
    """
    The main function to parse command-line arguments and initiate the replay of chat messages.
    """
    parser = argparse.ArgumentParser(
        description='Replay chat messages from a JSON file.')
    parser.add_argument('input_file',
                        metavar='FILE',
                        help='JSON file containing the chat messages')
    parser.add_argument(
        '--render',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Render assistant messages with rich Markdown (default: True)')
    args = parser.parse_args(argv)
    replay(args.input_file, args.render)


if __name__ == '__main__':  # pragma: no cover
    main()



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import Union, List
import requests
from .defaults import CACHE
from .cache import Cache


class DebianPolicy:
    '''
    Cache the plain text policy document and query its sections / subsections.
    '''
    NAME: str = 'Debian Policy'
    URL: str = 'https://www.debian.org/doc/debian-policy/policy.txt'
    SEP_SECTION: str = '***'
    SEP_SUBSECTION: str = '==='
    SEP_SUBSUBSECTION: str = '---'

    def __init__(self) -> None:
        cache = Cache(CACHE)
        # Check if the cache exists and read lines
        if self.URL not in cache:  # pragma: no cover
            r = requests.get(self.URL)
            cache[self.URL] = r.text
            lines = r.text.split('\n')
        else:
            lines = cache[self.URL].split('\n')

        # Scan the document and prepare the section indexes.
        self.lines: List[str] = [x.rstrip() for x in lines]
        self.indexes: list[str] = self.__scan_indexes()

    def __iter__(self):
        # Return an iterator over the section indexes.
        self.__cursor: int = 0
        return self

    def __next__(self) -> str:
        # Return the next section index.
        if self.__cursor < len(self.indexes):
            section = self.indexes[self.__cursor]
            self.__cursor += 1
            return self.__getitem__(section)
        else:
            raise StopIteration

    def __len__(self) -> int:
        # Return the number of sections in the document.
        return len(self.indexes)

    def __scan_indexes(self) -> list[str]:
        # Scan the document and return a list of all section indexes.
        ret: list[str] = []
        for i in range(1, len(self.lines)):
            cursur = self.lines[i]
            previous = self.lines[i - 1]
            if any(
                    cursur.startswith(x) for x in [
                        self.SEP_SECTION, self.SEP_SUBSECTION,
                        self.SEP_SUBSUBSECTION
                    ]):
                index = previous.split(' ')[0]
                if index.endswith('.'):
                    ret.append(index.rstrip('.'))
        return ret

    def __str__(self) -> str:
        # Return the entire document as a string.
        return '\n'.join(self.lines)

    def __getitem__(self, index: Union[str, int]) -> str:
        # if the index is an integer, map it to the real section number
        if isinstance(index, int):
            section = self.indexes[index]
            return self.__getitem__(section)
        # Retrieve a specific section, subsection, or subsubsection based on the index.
        sep: str = {
            1: self.SEP_SECTION,
            2: self.SEP_SUBSECTION,
            3: self.SEP_SUBSUBSECTION
        }[len(index.split('.'))]

        ret: list[str] = []
        prev: str = ''
        in_range: bool = False

        # Iterate over lines to find the specified section.
        for cursor in self.lines:
            if cursor.startswith(sep) and prev.startswith(f'{index}. '):
                # Start of the desired section
                ret.append(prev)
                ret.append(cursor)
                in_range = True
            elif cursor.startswith(sep) and in_range:
                # End of the desired section
                ret.pop(-1)
                in_range = False
                break
            elif in_range:
                # Within the desired section
                ret.append(cursor)
            prev = cursor

        return '\n'.join(ret)


class DebianDevref(DebianPolicy):
    NAME: str = "Debian Developer's Reference"
    URL: str = 'https://www.debian.org/doc/manuals/developers-reference/developers-reference.en.txt'


if __name__ == '__main__':  # pragma: no cover
    # Test the DebianPolicy class.
    p = DebianPolicy()
    print('Policy total length', len(str(p).encode()), 'bytes')
    for (sec, text) in zip(p.indexes, p):
        print('section', sec, 'length', len(text.encode()), 'bytes')

    # Test the DebianDevref class.
    d = DebianDevref()
    print('Devref total length', len(str(d).encode()), 'bytes')
    for (sec, text) in zip(d.indexes, d):
        print('section', sec, 'length', len(text.encode()), 'bytes')



'''
Copyright (C) 2024-2025 Mo Zhou <lumin@debian.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from types import SimpleNamespace
import sys
import os
import numpy as np
import pytest
from debgpt import defaults
from debgpt import embeddings


@pytest.fixture
def conf() -> object:
    return defaults.Config()


def test_random_embedding_embed(conf):
    model = embeddings.RandomEmbedding(conf)
    vector = model.embed('hello world')
    assert vector.ndim == 1
    # test __call__
    emb = model('hello world')
    assert emb.ndim == 1
    assert np.isclose(np.linalg.norm(emb), 1.0)


def test_random_embedding_batch_embed(conf):
    model = embeddings.RandomEmbedding(conf)
    matrix = model.batch_embed(['hello world', 'goodbye world'])
    assert matrix.ndim == 2
    # test __call__
    emb = model(['hello world', 'goodbye world'])
    assert emb.ndim == 2
    assert np.isclose(np.linalg.norm(emb, axis=1), 1.0).all()


def test_openai_embedding_embed(conf):
    if conf.openai_api_key == 'your-openai-api-key':
        pytest.skip('OpenAI API key is not provided')
    model = embeddings.OpenAIEmbedding(conf)
    vector = model.embed('hello world')
    assert vector.ndim == 1
    print(f'vector.shape:', vector.shape)
    print(f'vector.min:', vector.min())
    print(f'vector.max:', vector.max())
    print(f'vector.mean:', vector.mean())
    print(f'vector.std:', vector.std())
    print(f'vector[:10]:', vector[:10])
    print(f'vector[-10:]:', vector[-10:])

    # test __call__
    emb = model('hello world')
    assert emb.ndim == 1
    assert np.isclose(np.linalg.norm(emb), 1.0)


def test_openai_embedding_batch_embed(conf):
    if conf.openai_api_key == 'your-openai-api-key':
        pytest.skip('OpenAI API key is not provided')
    model = embeddings.OpenAIEmbedding(conf)
    matrix = model.batch_embed(['hello world', 'goodbye world'])
    assert matrix.ndim == 2
    print(f'matrix.shape:', matrix.shape)
    print(f'matrix.min:', matrix.min())
    print(f'matrix.max:', matrix.max())
    print(f'matrix.mean:', matrix.mean())
    print(f'matrix.std:', matrix.std())
    print(f'matrix[:, :10]:', matrix[:, :10])
    print(f'matrix[:, -10:]:', matrix[:, -10:])

    # test __call__
    emb = model(['hello world', 'goodbye world'])
    assert emb.ndim == 2
    assert np.isclose(np.linalg.norm(emb, axis=1), 1.0).all()


def test_google_embedding_embed(conf):
    if conf.google_api_key == 'your-google-api-key':
        pytest.skip('Google API key is not provided')
    model = embeddings.GoogleEmbedding(conf)
    vector = model.embed('hello world')
    assert vector.ndim == 1
    assert np.isclose(np.linalg.norm(vector), 1.0)


def test_google_embedding_batch_embed(conf):
    if conf.google_api_key == 'your-google-api-key':
        pytest.skip('Google API key is not provided')
    model = embeddings.GoogleEmbedding(conf)
    matrix = model.batch_embed(['hello world', 'goodbye world'])
    assert matrix.ndim == 2
    assert np.isclose(np.linalg.norm(matrix, axis=1), 1.0).all()


def test_get_embedding_model(conf):
    model = embeddings.get_embedding_model(conf)
    assert model is not None
    vector = model.embed('hello world')
    assert vector.ndim == 1
    assert np.isclose(np.linalg.norm(vector), 1.0)


def test_embedding_main():
    embeddings.main(['hello world'])
